/* 
 * Copyright (c) 2008-2020 ISP RAS
 *
 * Ivannikov Institute for System Programming of Russian Academy of Sciences
 * 25 A.Solzhenitsyn st. Moscow 109004 Russia
 *
 * All rights reserved.
 *
 * This source code is free software; you can redistribute it
 * and/or modify it in source code form under the terms of the GNU
 * General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 * 
 * $Id: veritool.c,v 1.5 2010/05/25 11:52:34 chupilko Exp $
 */

#include <assert.h>
#include <inttypes.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <ivl_target.h>

#include "design.h"
#include "error.h"
#include "options.h"
#include "printer.h"
#include "veritool.h"

vt_design_t *design = NULL;
int warnings = 0;

const char* reserved_names[] = {"and", "and_eq", "asm", "auto", "bitand", "bitor", "bool", "break",
"case", "catch", "char", "class", "continue", "default", "delete", "do", "double", "dynamic_cast",
"else", "enum", "emit", "extern", "false", "float", "for", "goto", "if", "incr", "int",
"long", "mutable", "new", "not", "not_eq", "register", "reinterpret_cast", "return",
"signed", "sizeof", "short", "static", "static_cast", "struct", "switch",
"template", "throw", "true", "try", "typedef", "typeid", "typename",
"union", "unsigned", "using", "void", "volatile", "wchar_t", "while", "xor", "xor_eq",
"begin", "end"};
#define RESERVED_NAMES_NUMBER (sizeof(reserved_names) / sizeof(char*))

// return 1 if the wire_name should be masked by a backslash and whitespace
// regular identifiers: start with a letter or underscore "_",
// contain letters, numbers, "_" and dollar sign "$",
// may contain any non-whitespace escaped character
tbm_t should_be_masked(const char *wire_name)
{
  assert(wire_name != 0);
  int i;

  for(i = 0; i < RESERVED_NAMES_NUMBER; i++)
    if(!strcmp(reserved_names[i], wire_name))
    {
      return TBM_RESERVED;
    }

  char s = wire_name[0];
  if(s == '$' || s >= '0' && s <= '9')
  {
    return TBM_ILLEGAL;
  }

  i = 0;
  while((s = wire_name[i++]) != '\0')
  {
    if(!(s >= 'a' && s <= 'z' || s >= 'A' && s <= 'Z' || s >= '0' && s <= '9' || s == '$' || s == '_'))
    {
      return TBM_ILLEGAL;
    }
  }

  return TBM_NONE;
}


static void process_signal(ivl_signal_t net, vt_options_t *options)
{
    const char *name = ivl_signal_basename(net);

    int inv_borders = ivl_signal_lsb(net) > ivl_signal_msb(net) ? 1 : 0;

    char *vname = malloc(strlen(name) + 2);
    char *vname_reg = malloc(strlen(name) + 6);
    char *cname = malloc(strlen(name) + 1);
    switch(should_be_masked(name))
    {
      case TBM_NONE:
        strcpy(vname, name);
        strcpy(vname_reg, vname); strcpy(vname_reg + strlen(name), "_reg");
        strcpy(cname, name);
      break;
      case TBM_RESERVED:
        strcpy(vname + 1, name); vname[0] = '\\'; strcpy(vname + strlen(name) + 1, " ");
        strcpy(vname_reg, vname); strcpy(vname_reg + strlen(name) + 1, "_reg ");
        strcpy(cname + 1, name); cname[0] = '_'; cname[strlen(name) + 1] = '\0';
      break;
      case TBM_ILLEGAL:
        strcpy(vname + 1, name); vname[0] = '\\'; strcpy(vname + strlen(name) + 1, " ");
        strcpy(vname_reg, vname); strcpy(vname_reg + strlen(name) + 1, "_reg ");
	strcpy(cname, name); cname[strlen(name)] = '\0';
	for(int i = 1; i < strlen(cname); i++)
        {
          if(!(cname[i] >= 'a' && cname[i] <= 'z' || cname[i] >= 'A' && cname[i] <= 'Z' || cname[i] >= '0' && cname[i] <= '9' || cname[i] == '_'))
          {
            cname[i] = '_';
          }
        }
      break;
    }

    vt_wire_t *wire = vt_create_wire(name, VT_NONE, VT_DATA, inv_borders == 0 ? ivl_signal_lsb(net) : ivl_signal_msb(net),
                                                             inv_borders == 0 ? ivl_signal_msb(net) : ivl_signal_lsb(net),
							     vname, cname, vname_reg);

    if(wire->max - wire->min + 1 > 512)
        { vt_format_error_and_exit("[ERROR] wire %s is too large: [%d:%d]", options, wire->name, wire->max, wire->min); }

    if(!strcmp(name, options->clock_input))
        { wire->input = VT_CLOCK; }
    else if(!strcmp(name, options->reset_input))
        { wire->input = VT_RESET; }
    
    switch(ivl_signal_port(net))
    {
        case IVL_SIP_INPUT:  { wire->direct = VT_INPUT;  break; }
        case IVL_SIP_OUTPUT: { wire->direct = VT_OUTPUT; break; }
        case IVL_SIP_INOUT:  { wire->direct = VT_INOUT;  break; }
        case IVL_SIP_NONE:   { wire->direct = VT_NONE;   break; }
    }

    if(wire->direct == VT_INPUT)
    {
      if(options->group && vt_find_and_update_similar_input_wire_in_design(design, wire))
      {
      }
      else
      {
        vt_add_input_design(design, wire);
      }
    }
    else if(wire->direct == VT_OUTPUT)
    {
      if(options->group && vt_find_and_update_similar_output_wire_in_design(design, wire))
      {
      }
      else
      {
        vt_add_output_design(design, wire);
      }
    }
    else if(wire->direct == VT_INOUT)
    {
      printf("[WARNING] INOUT-wires are not supported (%s is of such a type)\n", wire->name);
      warnings++;
    }
    else if(wire->direct == VT_NONE)
    {
      printf("[INFO] NONE-wire called \"%s\" is passed by during processing\n", wire->name);
    }

#ifdef DEBUG
    printf("[DEBUG] Wire %s (Verilog name: %s, reg_name: %s, C name: %s, type: %d) is under examination\n", wire->name, wire->vname, wire->vname_reg, wire->cname, wire->direct);
#endif
}

static void process_scope(ivl_scope_t net, vt_options_t *options)
{
    int i;

    design = vt_create_design(ivl_scope_name(net));

    for(i = 0; i < ivl_scope_sigs(net); i++)
        { process_signal(ivl_scope_sig(net, i), options); }

    if(options->group)
      vt_reduce_grouped_wire_names(design);
        
    vt_print_files(design, options);
    vt_destroy_design(design);
    
    printf("[OK] Scope %s has been analysed!\n", ivl_scope_name(net));
}

int target_design(ivl_design_t des)
{
    unsigned i, n;
    ivl_scope_t *scopes;

    printf("\nVeriTool started working...\n");
    
    vt_process_options(des, &options);

    ivl_design_roots(des, &scopes, &n);
    for(i = 0; i < n; i++)
    {
    	if(vt_defined_option(options.module_name) && strcmp(ivl_scope_name(scopes[i]), options.module_name))
    	    { continue; }
    	
    	// Process at most one Verilog-module.
    	process_scope(scopes[i], &options);
    	break;
    }
    
    if(i == n)
        { vt_print_error_and_exit("can't find a module", &options); }
    else
        { if(warnings == 0) { printf("[OK] VeriTool has finished its work successfully!\n\n"); }
	  else              { printf("[WARNING] VeriTool has finished its work but some problems have happened!\n\n"); }
	}

    return 0;
}
